//=============================================================================
//
//  CLASS Raytracer - IMPLEMENTATION
//
//=============================================================================
#define RAYTRACER_WHITTED_CPP

//== INCLUDES =================================================================
#include <QMainWindow>
#include <QProgressBar>
#include "Raytracer.h"


//== IMPLEMENTATION ==========================================================
//Die Hauptfunktion des Raytracers: Hier werden die Strahlen erzeugt
void Raytracer::render(
    SurfaceList     *pclScene, //Die Geometrie der Szene 
    unsigned int    uiMaxDepth, //Maximale Rekursionstiefe
    QImage          *clQImage, //das fertige Bild
    RenderingMode   mode //gibt an, ob der Raytracer im Timing-Modus oder im Anzeigen-Modus laeuft
)
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    ColorType   col( 0 );
    Ray         cl_ray;
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    //Inititialize progress Bar if in Timing Mode
    QProgressDialog progressDialog;
        progressDialog.setRange(0, clQImage->width() );
        progressDialog.setLabelText("Ray Tracing Progress");
        progressDialog.setModal(true);


    //Initialisierung der Kamera
    m_clPinholeCamera.setAspectRatio( clQImage->width() / (double) clQImage->height() );

    //Fr alle Pixel des Bildes
    for( int i = 0; i < clQImage->width(); i++ ) {
        progressDialog.setValue(i);
        if ( progressDialog.wasCanceled() )
           break;
        qApp->processEvents();

        for( int j = 0; j < clQImage->height(); j++ )
        {
            //OpenGL-Zeug, nicht wichtig zum Verstndnis der Raytracers
            //Deswegen bitte nicht weiter beachten
            //Wer`s wissen will: In glRays werden die Strahlen gespeichert, die
            // spter im OpenGL Fenster angezeigt werden wenn man "Show Rays" aktiviert
            std::vector<GLRay> *glRays; 
            if ( mode == MODE_DISPLAY )
            {
                glRays = new std::vector<GLRay>;
                glRays->clear();
            }
            else
                glRays = NULL;

            //Generiere Strahl
            Ray ray = getCamera().generateRay( i, j, clQImage->width(), clQImage->height() );

            //verfolge Strahl in die Szene und erhalte Farbe
            col = traceRay( pclScene, ray, uiMaxDepth, glRays );

            //Umrechnen der Farbe in Qt-Farben
            col.Clamp( 0, 1 );
            //Setzen des Pixels im Qt-Image
            clQImage->setPixel( i, j, col.getPacked() );

			//Anzeigen der Strahlen und des Bildes wenn Timing deaktiviert
			if ( mode == MODE_DISPLAY )
			{
                emit    renderImageAndRays( glRays );
				delete  glRays;
            }
        }
    }
}

//Verfolgung der Strahlen durch die Szene
ColorType Raytracer::traceRay(
    const Surface       *pclScene,
    const Ray           &crclRay,
    unsigned int        uiDepth,
    std::vector<GLRay>  *glRays )
{
    ColorType   col( 0 );

    //Maximale Rekursionstiefe erreicht?
    if( 0 == uiDepth )
        return col;
    --uiDepth;

    /*
        Implement:

        (i) if the surface that was hit is specular (Shader.isSpecular()), calculate 
			and trace the reflected ray
        (ii) if the hit surface is transparent (Shader.IsTransparent()), calculate and 
			trace the refracted ray
        (ii) add the weighted (Shader.getSpecularColor(), Shader.getRefractiveColor()) 
			colors of the secondary rays to the results of the local lighting model
    */

    //Struct zum speichern wichtiger Informationen: Siehe Types.h
    TracingContext context;

    //Berechnen des nchsten Schnittpunktes mit der Szene
    if ( pclScene->intersect(crclRay, SELF_ISECT_OFFSET, FLT_MAX, context) ) {
        context.m_v3Outgoing = -crclRay.getDir();
        context.m_v3Outgoing.normalize();

        //context.m_pclShader enthlt nach der Schnittpunktberechnung die Phong-Parameter fr das aktuell nchste Objekt
        const Shader* shader = context.m_pclShader;
        //Setze Phong-Beleuchtung
        col = shade(pclScene, context);

        //trace reflected ray
        if (shader->isSpecular()) {
            //Implementieren Sie:
            //Falls das Objekt einen spekularen Anteil hat: Berechne rekursiv den Farbanteil des gespiegelten Strahls
            //Verwende Ray::reflectedRay in Ray.cpp
        }
        //
        if (shader->isRefractive()) {
            //Implementieren Sie:
            //Falls das Objekt einen gebrochenen Anteil hat: Berechne rekursiv den Farbanteil des gebrochenen Strahls
            //Verwende Ray::refractedRay in Ray.cpp
        }


        //Setzen der Strahlen fr das OpenGL-Fenster
        //Nicht weiter wichtig zum Verstndnis des Raytracers
        if ( glRays != NULL ) {
            GLRay   glray;
            glray.m_start = crclRay.getOrigin();
            glray.m_stop = context.m_v3HitPoint;
            glRays->push_back( glray );
        }

        //Gib die berechnete Farbe zurck
        return col;
    } else {
        //Setzen der Strahlen fr das OpenGL-Fenster
        //Nicht weiter wichtig zum Verstndnis des Raytracers
        if ( glRays != NULL ) {
            GLRay   glray;
            glray.m_start = crclRay.getOrigin();
            glray.m_stop = crclRay.getOrigin() + crclRay.getDir() * 10000;
            glRays->push_back( glray );
        }
        //Falls der Strahl die Szene nicht schneidet: Gib die Hintergrundfarbe zurck
        return m_colBackground;
    }
}

//Implementiert das Phong-Beleuchtungsmodell
ColorType Raytracer::shade(
    const Surface   *pclScene,
    TracingContext  &tContext )
{
    /*
        Das lokale Beleuchtungsmodell ist durch von Shader abgeleitete
        Klassen definiert. Interessant sind hier die virtuellen Methoden
        shadeAmbient() und shade(). Der aktuelle Shader befindet sich
        im Argument tContext.m_
        Die Lichtquellen befinden sich im STL-Container m_LightList.
        Lichtquellen sind von Light abgeleitet und implementieren
        die Methode shadowFeeler(), welche Strahlrichtung, -Gltigkeitsbereich
        und -Intensitt zurckliefert.
        Die berprfung Schnitt Strahl - Szene kann ber die Methode
        Surface scene->intersect(...) abgefragt werden. Allgemein ist sie
        in Surface.h beschrieben, von welcher jedes Objekt abgeleitet wird.
    */
    //Setze den ambienten Term der Farbe
    ColorType color(tContext.m_pclShader->getAmbientColor());

    //fr alle Lichtquellen 
    LightListType::const_iterator it = m_LightList.begin();		
    LightListType::const_iterator it_end = m_LightList.end();		
    for( ; it != it_end; ++it ) {
        //Implementieren Sie:

        //berechne Schattenstrahl und Intensitt ( Siehe shadowFeeler() in Lights.h )
        //schneide Schattenstrahl mit der Szene
        //falls kein Schnittpunkt zwischen Objekt und Lichtquelle:
        //Auswertung des lokalen Beleuchtungsmodells (Shader.shade()) 
        //Addiere das Ergebnis zur aktuellen Farbe
    }
    return color;
}

//Zurcksetzen der Lichtquellen
void Raytracer::removeAllLights()
{
    if( m_bOwnsLights )
    {
        LightListType::iterator it = m_LightList.begin();
        LightListType::iterator it_end = m_LightList.end();
        for( ; it != it_end; ++it )
        {
            delete( *it );
        }
    }
    m_LightList.clear();
}


//Zeichnen der Lichtquellen in OpenGL: nicht weiter wichtig fr den Raytracer
void Raytracer::renderLightsGL()
{
    LightListType::iterator it = m_LightList.begin();
    LightListType::iterator it_end = m_LightList.end();
    glEnable( GL_LIGHTING );
    for( unsigned int lightcounter = 0; it != it_end; ++it )
    {
        GLint   li;
        switch( lightcounter )
        {
            case 0:     li = GL_LIGHT0; break;
            case 1:     li = GL_LIGHT1; break;
            case 2:     li = GL_LIGHT2; break;
            case 3:     li = GL_LIGHT3; break;
            case 4:     li = GL_LIGHT4; break;
            case 5:     li = GL_LIGHT5; break;
            case 6:     li = GL_LIGHT6; break;
            case 7:     li = GL_LIGHT7; break;
            default:    li = GL_LIGHT7; break;
        }
        ( *it )->renderGL( li );
        glEnable( li );
        lightcounter++;
    }
}
