//=============================================================================
//
//  Raytracing Widget
//
//=============================================================================
#ifndef RENDERAREA_H
#define RENDERAREA_H
#include <QImage>
#include <QWidget>
#include <QPainter>
#include "Raytracer.h"
#include "SurfaceList.h"
#include "PhongShader.h"
#include "Types.h"
class RaytracingArea : public QWidget
{
    Q_OBJECT
    
public:
    enum Shape { Image };
    RaytracingArea( QWidget *parent = 0 );
    QSize   minimumSizeHint() const;
    QSize   sizeHint() const;
public slots :
	void	openScene ( const QString &path = QString ( ) );
    void    saveImage( const QString &path = QString() );
    void    render();
    void    renderImage( RayArray * glRays );
	void	updateImageSize( int size );
	void	updateRecursionDepth( int depth );
	void	setDisplayMode( int mode );
signals:
    void    setGLScene( Raytracer * raytracer,
                        SurfaceList * scene );
    void    renderRaysGL( RayArray * glRays );
    void    resetGL();
	void	startRaytracingGL( QImage *textureImage );


protected:
    void    paintEvent( QPaintEvent *event );


private:
    QProgressBar    *progressBar;
    Raytracer *m_raytracer;
    SurfaceList *m_scene;
    std::vector<Shader *>   *m_materials;
    QImage                  *m_raytracedImage;
	RenderingMode	m_displayMode;
	int m_recursionDepth;
	int m_imageSizeX;
	int m_imageSizeY;

};
#endif
